""" Components are hardware-defined components where a stimulus can be
applied or an input can be read. Defining components is the basis of
automated test coverage calculation.
"""

import collections
from .nets import first_alias, flatten_string_list
from ..util import yaml_loader

QUALIFIED_NAME_SEPARATOR = '.'

def qualified_name_join(*strings):
    return QUALIFIED_NAME_SEPARATOR.join(strings)

def net_to_fully_qualified_name(prefix, net):
    return qualified_name_join(prefix, first_alias(net))

def load_component_file(filename):
    content = yaml_loader.load_yml_file(filename)
    return Component(**content)
    
class Component(object):
    def __init__(self, name="", description="", components=[], nets=[], refdes=[], **kwargs):
        self._parent = None
        self.name = name
        self.description = description
        
        self.components = [None]*len(components)
        for i, component in enumerate(components):
            if isinstance(component, collections.Mapping):
                component = self.__class__(**component)
            
            component.attach_to(self)
            self.components[i] = component
        
        self.nets = flatten_string_list(nets)
        self.refdes = refdes
        self._nets_parsed = False
        
    def attach_to(self, parent):
        self._parent = parent
        
    def fully_qualified_prefix(self):
        if self._parent:
            prefix = self._parent.fully_qualified_prefix()
            if prefix:
                return qualified_name_join(prefix, self.name)
            else:
                return self.name
        else:
            return self.name
        
    def fully_qualified_nets(self, prefix=None):
        if prefix is None:
            prefix = self.fully_qualified_prefix()
        return [net_to_fully_qualified_name(prefix, net) for net in self.nets]
        
    def iter_all(self, include_self=True):
        if include_self:
            yield self
        
        for top_component in self:
            for component in top_component.iter_all(include_self=True):
                yield component
    
    def iter_fully_qualified_nets(self):
        nets = []
        for component in self.iter_all():
            nets += component.fully_qualified_nets()
        return nets
        
    def __iter__(self):
        for component in self.components:
            yield component
    
    


import pytest
import time
import threading
import queue

from spintop.plugs.comport import ComportInterface

class FakeSerial(object):
    def __init__(self):
        self.to_read = ""
        self.is_open = True
        self.in_waiting = False
        self.echo = True
        self.was_read = threading.Event()

    def close(self):
        self.is_open = False

    def read(self, *args, **kwargs):
        if self.to_read:
            value = self.to_read
            self.to_read = ""
            self.in_waiting = False
            self.was_read.set()
            return value.encode('utf8')
        else:
            time.sleep(0.5)
            return None
    
    def write(self, string):
        if self.echo:
            self.in_waiting = True
            self.to_read += string.decode('utf8')

@pytest.fixture()
def fake_serial():
    fake_serial = FakeSerial()
    return fake_serial

@pytest.fixture()
def comport(fake_serial):
    # comport = ComportInterface('/dev/serial/by-id/usb-STMicroelectronics_STM32_Virtual_ComPort_in_FS_Mode_00000000050C-if00', 115200)
    comport = ComportInterface(None, None)
    comport.open(fake_serial)
    yield comport
    comport.close()


def test_receive_no_eol(fake_serial, comport):
    comport.write('Hello')
    line = comport.next_line(timeout=2)
    assert line == 'Hello'

def test_receive_immediate_eol(fake_serial, comport):
    fake_serial.was_read.clear()
    comport.write('Hello\n')
    # no wait now (only enough time for the thread to call read)
    fake_serial.was_read.wait()
    assert comport.next_line(timeout=0.1) == 'Hello\n'


import os
import sys
import inspect
from copy import copy
from spintop.storage import SITE_DATA_DIR

import openhtf as htf

from openhtf.util import conf
from openhtf.plugs import user_input, BasePlug
from openhtf.output.callbacks import json_factory

import webbrowser

from ..callbacks import station_server

from .. import (
    Test,
    # load_component_file,
    # CoverageAnalysis
)

HISTORY_BASE_PATH = os.path.join(SITE_DATA_DIR, 'openhtf-history')

DEFAULT = object()

class TestPlanError(Exception): pass

class DecorativeTestNode(object):
    def __init__(self, name):
        self._setup_phases = []
        self._test_phases = []
        self._teardown_phases = []
        self.name = name
    
    def setup(self, name):
        return self._decorate_phase(name, self._setup_phases)
    
    def testcase(self, name, tests=None, targets=[]):
        # if tests and self._top_level_component is None:
        #     raise TestPlanError('The top level component must be defined using the define_top_level_component function of TestPlan in order to use tests or targets coverage parameters.')
        
        # def _note_fn(fn):
        #     phase = self._add_phase(fn, name, self._test_phases)
        #     # if tests:
        #     #     print(self.coverage.add_test(tests, name, allow_links_to=targets))
            
        #     return phase
        return self._decorate_phase(name, self._test_phases)
    
    def teardown(self, name):
        return self._decorate_phase(name, self._teardown_phases)
    
    def plug(self, *args, **kwargs):
        """Helper method: shortcut to htf.plugs.plug(...)"""
        return htf.plugs.plug(*args, **kwargs)
    
    def sub_group(self, name):
        group = DecorativeTestNode(name)
        self._test_phases.append(group)
        return group
    
    def _decorate_phase(self, name, array):
        def _note_fn(fn):
            phase = self._add_phase(fn, name, array)
            return phase
        return _note_fn
    
    def _add_phase(self, fn, name, array):
        phase = ensure_htf_phase(fn)
        phase.options.name = name
        # phase.extra_kwargs['testplan'] = self
        array.append(phase)
        return phase
        
    @property
    def phase_group(self):
        # Recursively get phase groups of sub phases if available, else the phase itself.
        _test_phases = [getattr(phase, 'phase_group', phase) for phase in self._test_phases]
        
        return htf.PhaseGroup(
            setup=self._setup_phases,
            main=_test_phases,
            teardown=self._teardown_phases,
            name=self.name
        )
    
class TestPlan(DecorativeTestNode):
    def __init__(self, name='testplan'):
        super(TestPlan, self).__init__(name=name)
        self._top_level_component = None
        self.coverage = None
        self.file_provider = station_server.TemporaryFileProvider()
        self.callbacks = []
        
        # Array but must contain only one phase.
        # Array is for compatibility with self._decorate_phase function of parent class.
        self._trigger_phases = []
        
        self.add_callbacks(json_factory.OutputToJSON(
            os.path.join(HISTORY_BASE_PATH,'{metadata[test_name]}', '{dut_id}-{start_time_millis}-{outcome}.json'), indent=4))

    @property
    def history_path(self):
        path = os.path.join(HISTORY_BASE_PATH, self.name)
        if not os.path.exists(path):
            os.makedirs(path)
        return path
    
    @property
    def trigger_phase(self):
        return self._trigger_phases[0] if self._trigger_phases else None
    
    def trigger(self, name):
        if self.trigger_phase:
            raise TestPlanError('There can only be one @trigger function.')
        
        return self._decorate_phase(name, self._trigger_phases)

    def run(self, launch_browser=True, **execute_kwargs):
        with station_server.StationServer(self.file_provider) as server:
            self.add_callbacks(server.publish_final_state)
            self.configure()
            
            if launch_browser and conf['station_server_port']:
                webbrowser.open('http://localhost:%s' % conf['station_server_port'])
                
            while True:
                try:
                    self.execute(**execute_kwargs)
                except KeyboardInterrupt:
                    break
    
    def configure(self):
        self.test = Test(self.phase_group, test_name=self.name, _code_info_after_file=__file__)
        self.test.configure(failure_exceptions=(user_input.SecondaryOptionOccured,))
        self.test.add_output_callbacks(*self.callbacks)
        
    def add_callbacks(self, *callbacks):
        self.callbacks += callbacks
    
    def execute(self, test_start=DEFAULT):
        """ Execute the configured test using the test_start function as a trigger.
        """
        if test_start is DEFAULT:
            if self.trigger_phase is None:
                self._create_default_trigger()
            test_start = self.trigger_phase
            
        return self.test.execute(test_start=test_start)
    
    def create_plug(self):
        class _SelfReferingPlug(BasePlug):
            def __new__(cls):
                return self
        
        return _SelfReferingPlug
    
    def spintop_plug(self, fn):
        return htf.plugs.plug(spintop=self.create_plug())(fn)
    
    def _create_default_trigger(self, message='Enter a DUT ID in order to start the test.',
            validator=lambda sn: sn, **state):
        
        @self.trigger('Simple Scan')
        @htf.PhaseOptions(timeout_s=None, requires_state=True)
        @htf.plugs.plug(prompts=user_input.UserInput)
        def trigger_phase(state, prompts):
            """Test start trigger that prompts the user for a DUT ID."""
            dut_id = prompts.prompt(message, text_input=True)
            state.test_record.dut_id = validator(dut_id)
            
        return trigger_phase
    
    def define_top_level_component(self, _filename_or_component):
        if isinstance(_filename_or_component, str):
            component = load_component_file(_filename_or_component)
        else:
            component = _filename_or_component
        self._top_level_component = component
        self.coverage = CoverageAnalysis(self._top_level_component)

def ensure_htf_phase(fn):
    if not hasattr(fn, 'options'):
        # Not a htf phase, decorate it so it becomes one.
        fn = htf.PhaseOptions()(fn) 
    return fn



from json import loads, dumps
from openhtf.plugs.user_input import UserInput

class FormUserInput(UserInput):
    def prompt(self, message, form_schema, *args, **kwargs):
        response = super(FormUserInput, self).prompt(message=self.pack_message(message, form_schema), *args, **kwargs)
        if response:
            response = self.unpack_response(response)
        return response
        
    def pack_message(self, message, form_schema):
        return dumps({
            'message': message,
            'form_schema': form_schema
        })
        
    def unpack_response(self, response):
        return loads(response)


""" The main spintop entry point. """

from .core.test_descriptor import (
    Test
)

from .testplan.base import (
    TestPlan
)

# from .coverage.analysis import (
#     create_netlist_from_component,
#     create_netcomp_map_from_component,
#     CoverageAnalysis,
#     CoverageAnalysisError
# )

# from .coverage.components import (
#     Component,
#     load_component_file
# )

# from .coverage.nets import (
#     load_nets,
#     load_nets_yml,
#     Net
# )


from .standard import (
    EnvironmentType,
    get_env,
    is_development_env,
    is_production_env
)



import openhtf.plugs as plugs
from openhtf.util import conf

import time

import serial
from serial.threaded import ReaderThread

import threading
from queue import Queue
from asyncio import Protocol

class ComportInterface(Protocol):

    def __init__(self, comport, baudrate=115200):
        self.comport = comport
        self.baudrate = baudrate

        self._serial = None
        self._reader = None
        self._line_buffer = ""
        self._read_lines = Queue()
        self._timeout_timer = None
        self._last_receive_time = None

        self.timeout = 0.5
        self.eol = "\n"

    def tearDown(self):
        """Tear down the plug instance."""
        self.close()

    def close(self):
        if self._reader is not None and self._serial.is_open:
            self._reader.close()
            self._reader = None

    def open(self, _serial=None):
        self.close()
        if not _serial:
            _serial = serial.Serial(self.comport, self.baudrate, timeout=self.timeout)
        self._serial = _serial
        self._reader = ReaderThread(self._serial, lambda: self)
        self._reader.start()

    def write(self, data):
        return self._reader.write(data.encode('utf8'))

    def data_received(self, data):
        if self._timeout_timer is not None:
            self._timeout_timer.cancel()
        
        force_clear = False
        if self.check_receive_delta() and data is None and self._line_buffer:
            force_clear = True

        data_to_queue = None

        if not data:
            data = b''

        string = data.decode('utf8')
        if force_clear or string.endswith(self.eol) or (self._line_buffer + string).endswith(self.eol):
            data_to_queue = self._line_buffer + string
            self._line_buffer = ""
        else:
            self._line_buffer += string
        
        self._timeout_timer = threading.Timer(self.timeout, self.no_data_received)
        self._timeout_timer.start()

        if data_to_queue:
            self._read_lines.put(data_to_queue)

    def no_data_received(self):
        return self.data_received(None)

    def check_receive_delta(self):
        if self._last_receive_time:
            delta = time.time() - self._last_receive_time
        else:
            delta = 0

        self._last_receive_time = time.time()

        return delta > self.timeout

    def next_line(self, timeout=10):
        """ Waits up to timeout seconds and return the next line available in the buffer.
        """
        return self._read_lines.get(timeout=timeout)

    def eof_received(self):
        pass

def define_comport_plug(comport_conf_name, comport_conf_baudrate=None):
    conf.declare(comport_conf_name, description='Declared comport accessor name')

    if comport_conf_baudrate is not None:
        conf.declare(comport_conf_baudrate, description='Declared comport baudrate')

    class ComportPlug(ComportInterface, plugs.BasePlug):
        def __init__(self):
            if comport_conf_baudrate is None:
                baudrate = 115200
            else:
                baudrate = conf[comport_conf_baudrate]

            super(ComportPlug, self).__init__(conf[comport_conf_name], baudrate)

    return ComportPlug



"""
    If a result is present, the netcomp is covered
    If tested is True, then the result compliant defines if the netcomp is compliant or not.
    If tested is False, then the netcomp is covered but a test did not run.
"""
from fnmatch import fnmatchcase
from collections import namedtuple
from .nets import load_nets, first_alias
from .components import net_to_fully_qualified_name

TestResult = namedtuple('CoverageResult', 'name tested compliant')

NetComp = namedtuple('NetComp', 'key results net component')

class CoverageAnalysisError(Exception): pass

def create_netlist_from_component(top_level_component):
    nets = []
    for component in top_level_component.iter_all(include_self=True):
        nets += component.nets
    
    all_nets = load_nets(nets)
    return all_nets

def create_netcomp(key, net, component, netlist):
    net = netlist[first_alias(net)]
    netcomp = NetComp(key=key, results=[], net=net, component=component)
    # net.netcomps.append(netcomp) # back reference
    return netcomp
    

def create_netcomp_map_from_component(top_level_component, netlist):
    netcomps = {}
    for component in top_level_component.iter_all(include_self=True):
        prefix = component.fully_qualified_prefix()
        for net in component.nets:
            key = net_to_fully_qualified_name(prefix, net)
            netcomp = create_netcomp(key=key, net=net, component=component, netlist=netlist)
            netcomps[key] = netcomp
    return netcomps

def filter_netcomp_map(netcomp_map, must_match=None, allowed_nets=None):
    if must_match is None and allowed_nets is None:
        return netcomp_map
    
    def _nets_condition(netcomp_key):
        if netcomp_map[netcomp_key].net in allowed_nets:
            return True
        else:
            return False
    
    def _match_condition(netcomp_key):
        for pattern in must_match:
            if fnmatchcase(netcomp_key, pattern):
                    return True
        # else
        return False
    
    
    condition_fns = []
    if must_match is not None:
        condition_fns.append(_match_condition)
        
    if allowed_nets is not None:
        condition_fns.append(_nets_condition)
    
    def _filter(netcomp_key):
        for cond_fn in condition_fns:
            if not cond_fn(netcomp_key):
                # Return false as soon as one condition fails
                return False
        # else
        return True
        
    return filter(_filter, netcomp_map)

def create_subset_of_dict(_dict, keys):
    return {key:_dict[key] for key in keys if key in _dict}

class CoverageAnalysis(object):
    def __init__(self, top_level_component):
        self._top_level = top_level_component
        self._netlist = create_netlist_from_component(top_level_component)
        self._netcomp_map = create_netcomp_map_from_component(top_level_component, self._netlist)
        self._tests = {}
    
    def add_test(self, fnmatch, name, allow_links_to=[]):
        test = NetCompTestDescriptor(fnmatch, name, allow_links_to=allow_links_to)
        return test.filter_for_coverage(self._netcomp_map)
            

class NetCompTestDescriptor(object):
    def __init__(self, fnmatches, name, allow_links_to=[]):
        self.source_matches = fnmatches
        self.name = name
        self.targets_match_allow = allow_links_to
    
    def filter_for_coverage(self, netcomp_map):
        source_netcomp_keys = self.get_source_netcomp_keys(netcomp_map)
        
        source_nets = [netcomp_map[key].net for key in source_netcomp_keys]
        
        target_netcomp_keys = self.get_target_netcomp_keys(netcomp_map, source_nets)
        
        return set(target_netcomp_keys + source_netcomp_keys)
        
    def get_source_netcomp_keys(self, netcomp_map):
        return list(filter_netcomp_map(netcomp_map, must_match=self.source_matches))
    
    def get_target_netcomp_keys(self, netcomp_map, allowed_nets):
        return list(filter_netcomp_map(netcomp_map, must_match=self.targets_match_allow, allowed_nets=allowed_nets))
        
        
        
    




import pytest

from spintop.util import yaml_loader
from spintop.testplan import TestPlan, ElectricalAnalysisContext

def test_testplan_sequencing():
    plan = TestPlan()

    @plan.testcase('A')
    def fn1():
        pass

    @plan.testcase('B')
    def fn2():
        pass

    assert plan.phases == [fn1, fn2]


@pytest.fixture()
def fpga_yml_content():
    return yaml_loader.load_yml("""
nodes:
  - name: power
    nodes:
      - name: 5V
      - name: 3V
        implies: 
          - 5V
  - name: gps
    refdes: U33
    type: chip
    nodes:
      - name: uart
        testpoints:
          - GPS_UART
      - name: pps_out
        edges:
          - PPS_OUTPUT
    implies:
      - power.5V
  - name: uart_translator
    refdes: U231
    type: chip
    nodes:
      - name: uart
        edges:
          - GPS_UART
          - FPGA_UART
    implies:
      - power.5V
      - power.3V
  - name: fpga
    refdes: U4
    type: chip
    nodes:
      - name: uart
        edges:
          - FPGA_UART
      - name: pps_in
        edges:
          - PPS_OUTPUT
    implies:
      - power.3V
    """)
    
@pytest.fixture()
def fpga_yml_analysis(fpga_yml_content):
    analysis = ElectricalAnalysisContext()
    analysis.parse(fpga_yml_content)
    return analysis

def test_children_nodes(fpga_yml_content):
    
    
def test_nets_array():
    nets = [
        'A',
        'B',
        'C',
        'A, E' # Same Net
    ]

    nets = load_nets(nets)

    assert len(nets) == 4
    assert nets['A'] is nets['E'] # Assert same object
    
    assert 'A' in nets and 'E' in nets['A'].refs
    assert 'B' in nets
    assert 'C' in nets
    assert 'E' in nets and 'A' in nets['E'].refs

def test_nested_nets_array():
    nets = [
        'A',
        ['B', 'C']
    ]

    nets = load_nets(nets)
    assert len(nets) == 3
    for letter in 'ABC':
        assert letter in nets
    

def test_jinja2_expansion():
    net_yml_str = """
---
nets:
    {% for i in range(4) %}
    - A{{i}}
    {% endfor %}
    """

    nets = load_nets_yml(net_yml_str)

    assert len(nets) == 4
    for i in range(4):
        assert 'A%d' % i in nets
        
def test_components_to_netlist():
    """
    Defined nets: A, B, C
    Defined Components: X, Y, Z
    """
    top_level = top_two_child_components(first_nets=['A', 'B'], second_nets=['C'])
    
    netlist = create_netlist_from_component(top_level)
    
    assert len(netlist) == 3
    for net in 'ABC':
        assert net in netlist
        
def top_two_child_components(top_nets=[], first_nets=[], second_nets=[]):
    return Component(name='X', nets=top_nets, components=[
        Component(name='Y', nets=first_nets),
        Component(name='Z', nets=second_nets),
    ])

def four_level_deep_components(nets_per_component=([],[],[],[])):
    return \
    Component(name='W', nets=nets_per_component[0], components=[
        Component(name='X', nets=nets_per_component[1], components=[
            Component(name='Y', nets=nets_per_component[2], components=[
                Component(name='Z', nets=nets_per_component[3], components=[])
            ])
        ])
    ])
        
def test_deep_components_to_netlist():
    """
    Defined nets: A, B, C, D
    Defined Components: W, X, Y, Z
    """
    top_level = four_level_deep_components(nets_per_component=(
        ['A'], ['B'], ['C'], ['D']
    ))
    
    netlist = create_netlist_from_component(top_level)
    
    assert len(netlist) == 4
    for net in 'ABCD':
        assert net in netlist
        
def test_netlist_links_components_nets():
    """
    Defined nets: A, B, C
    Defined Components: W, X
    """
    top_level = top_two_child_components(first_nets=['A,B'], second_nets=['B,C'])
    
    netlist = create_netlist_from_component(top_level)
    
    assert len(netlist) == 3
    assert netlist['A'] is netlist['B']
    assert netlist['B'] is netlist['C']
    
def test_netlist_links_components_nets():
    """
    Defined nets: A, B, C
    Defined Components: W, X
    """
    top_level = top_two_child_components(first_nets=['A,B'], second_nets=['B,C'])
    
    netlist = create_netlist_from_component(top_level)
    
    assert len(netlist) == 3
    assert netlist['A'] is netlist['B']
    assert netlist['B'] is netlist['C']

def test_components_fully_qual_nets():
    """ Components are named X, Y(first_nets), Z(second_nets) """
    top_level = top_two_child_components(first_nets=['A', 'B'], second_nets=['C'])
    
    net_qual_names = top_level.iter_fully_qualified_nets()
    
    assert set(net_qual_names) == {'X.Y.A', 'X.Y.B', 'X.Z.C'}
    
def test_analysis_netcomp():
    """ Components are named X, Y(first_nets), Z(second_nets) """
    top_level = top_two_child_components(first_nets=['A'], second_nets=['A'])
    
    analysis = CoverageAnalysis(top_level)
    keys = analysis.add_test(['X.Y.*'], 'test_a', allow_links_to=[])
    
    assert 'X.Y.A' in keys
    assert 'X.Z.A' not in keys
    
    keys = analysis.add_test(['X.Y.*'], 'test_b', allow_links_to=['*'])
    
    assert len(keys) == 2
    assert 'X.Y.A' in keys
    assert 'X.Z.A' in keys


import pytest

from spintop.util import yaml_loader
from spintop.coverage import new_analysis

@pytest.fixture()
def fpga_yml_content():
    return yaml_loader.load_yml("""
components:
  - name: power
    components:
      - name: 5V
      - name: 3V
        implies: 
          - root.power.5V
  - name: gps
    refdes: U33
    type: chip
    components:
      - name: uart
        testpoints:
          - GPS_UART
      - name: pps_out
        testpoints:
          - PPS_OUTPUT
    implies:
      - root.power.5V
  - name: uart_translator
    refdes: U231
    type: chip
    components:
      - name: uart
        testpoints:
          - GPS_UART
          - FPGA_UART
    implies:
      - root.power.5V
      - root.power.3V
  - name: fpga
    refdes: U4
    type: chip
    components:
      - name: uart
        testpoints:
          - FPGA_UART
      - name: pps_in
        testpoints:
          - PPS_OUTPUT
    implies:
      - root.power.3V

interfaces:
  - name: uart_gps_fpga
    type: UART
    endpoints:
      - gps.uart # GPS_UART_RX, GPS_UART_TX on GPS chip side
      - fpga.uart # FPGA_UART_RX, FPGA_UART_TX on FPGA side
    intermediate:
      - uart_translator.uart # Both low and high UART sides on translator
  - name: gps_pps
    type: GPS_PPS
    enpoints:
      - fpga.pps_in
      - gps.pps_out
    """)

def test_parse(fpga_yml_content):
    
    analysis = new_analysis.NodeTreeContext()
    analysis.parse(fpga_yml_content)
    g = analysis.build_graph()

    from spintop.util.graph import auto_graph_to_plotly_fig
    
    fig = auto_graph_to_plotly_fig(g, seed=98)
    fig.show()
    


import itertools
import os

from jsonschema.validators import validator_for
from collections import Sequence

from ..util import yaml_loader
from ..util.compat import isstr

from .base import TestPlanError

HERE = os.path.abspath(os.path.dirname(__file__))

class AnalysisError(TestPlanError): pass

def flatten_string_list(array):
    new_array = []
    for item in array:
        if isinstance(item, Sequence) and not isinstance(item, str):
            new_array += item
        else:
            new_array.append(item)
            
    return new_array

def get_schema(filename=os.path.join(HERE, 'analysis_schema.yml')):
    return yaml_loader.load_yml_file(filename)

def compile_schema():
    schema = get_schema()
    Validator = validator_for(schema)
    Validator.check_schema(schema)
    return Validator(schema)

class DomainNodeInterface(object):
    def __init__(self, obj):
        self._dict = obj
    
    @property
    def name(self):
        return self._dict.get('name', '')
    
    @property
    def description(self):
        return self._dict.get('description', '')
    
    @property
    def nodes(self):
        return self._dict.get('nodes', [])
    
    @property
    def edges(self):
        return self._dict.get('edges', [])
    
    @property
    def required(self):
        return self._dict.get('required', False)
    
    @property
    def implies(self):
        return self._dict.get('implies', [])
    
    @property
    def testable(self):
        return self._dict.get('testable', False)

    @property
    def data(self):
        """ Domain specific data. """
        return {}
    
    def hook_node_attach_to(self, this_node, parent_node):
        """ Called when the node created from this interface, this_node, is attached
        to a parent_node """
        
    def hook_child_attach(self, this_node, child_node):
        """ Called when a child attached to the node created from this interface. """
    
    def as_node(self, context):
        
        node_objects = [context.to_node(child_node) for child_node in self.nodes]
        
        node = Node(
            name=self.name,
            description=self.description,
            nodes=node_objects,
            edges=self.edges,
            required=self.required,
            implies=self.implies,
            testable=self.testable,
            data=self.data
        )
        
        return node
    
class ElectricalDomainInterface(DomainNodeInterface):
    
    @property
    def nodes(self):
        return self._dict.get('components', []) + self._dict.get('subcomponents', [])
    
    @property
    def edges(self):
        return self._dict.get('testpoints', [])
    
    @property
    def testable(self):
        return False
    
    # def hook_node_attach_to(self, this_node, parent_node):

def ElectricalAnalysisContext():
    return AnalysisContext(domain_interface=ElectricalDomainInterface)

class AnalysisContext(object):
    def __init__(self, domain_interface=DomainNodeInterface):
        self.domain_interface = domain_interface
        self.root = Node(description="Root node")
        self.graph = None
        self.validator = compile_schema()
        
    def parse(self, node_dict):
        try:
            self.validator.validate(node_dict)
        except Exception as e:
            print(e)
        
        sub_root = self.to_node(node_dict)
        sub_root.attach_to(self.root)
        return sub_root
    
    def to_node(self, node_dict):
        return self.domain_interface(node_dict).as_node(self)
    
    def build_graph(self):
        """ Builds the dependency graph.
        
        The build is separated into two steps:
        1. Implied dependencies.
            These are dependencies specified by the shape of the tree and the parent/child
            relationships. When the 'required' node property is false, a compliant child
            adds an inbound test edge. When required is true, the the relationship is 
            inversed and the test coverage flows from the parent to the child.
        2. Direct dependencies.
            These are dependencies linked accross the tree using the defined edges.
        """
        import networkx as nx
        
        self.graph = nx.DiGraph()
        edges_map = {}
        self.nodes_map = {}
        
        if self.root is None:
            raise RuntimeError('self.parse must be called before build_connectivity_graph')
        
        # Step 1: Link nodes implied dependencies
        for node in self.root.iter_children():
            qualname = self._add_node(node)
            
            # Add nodes and edges to temporary maps
            for edge in node.edges:
                linked_nodes = edges_map.get(edge, set())
                linked_nodes.add(qualname)
                edges_map[edge] = linked_nodes
                
            # Create the implied dependencies
            if node.parent:
                if node.required:
                    self._add_edge(node.parent.qualname, qualname)
                else:
                    self._add_edge(qualname, node.parent.qualname)
            elif node.required:
                raise RuntimeError('Cannot have required property without a parent node.')
            
            for implied_qualname in node.implies:
                implied_ref = self.resolve_node_ref(node, implied_qualname)
                self._add_edge(qualname, implied_ref)
        
        # Step 2
        for linkname, nodes in edges_map.items():
            for u,v in itertools.combinations(nodes, 2):
                self._add_testable_edge(linkname, u, v)
        
        return self.graph
    
    def resolve_node_ref(self, node, ref, graph=None):
        if graph is None: graph = self.graph
        orig_ref = ref
        possible_refs = node.get_ref_qualname_possibilities(orig_ref)
        for ref in possible_refs:
            if ref in self.graph.nodes:
                return ref
        else:
            raise AnalysisError(
                'Unale to resolve ref "{ref}" from node "{node_qualname}". Tested possibilities: {possible_refs}'.format(
                    ref=orig_ref,
                    node_qualname=node.qualname,
                    possible_refs=possible_refs
                )
            )
        
    
    def _add_testable_edge(self, linkname, qualname_u, qualname_v):
        new_node = Node(
            name=qualname_u + '/' + linkname + '/' + qualname_v,
            testable=True
        )
        self._add_node(new_node)
        
        source = new_node.qualname
        self._add_edge(source, qualname_u)
        self._add_edge(source, qualname_v)
        
    
    def _add_node(self, node):
        qualname = node.qualname
        print('Add node %s' % qualname)
        self.graph.add_node(qualname, qualname=qualname, node=node)
        self.nodes_map[qualname] = node
        return qualname
        
    def _add_edge(self, u, v):
        print('Add edge %s to %s' % (u, v))
        self.graph.add_edge(u, v)
        

class Node(object):
    def __init__(self, name="", description="", nodes=[], edges=[], required=False, implies=[], testable=False, data={}, creator_interface=None):
        self._parent = None
        self._creator_interface = creator_interface
        
        self.name = name
        self.description = description
        self.edges = flatten_string_list(edges)
        self.required = required
        self.implies = implies
        self.testable = testable
        
        self.data = data
        
        self.nodes = []
        for node in nodes:
            node.attach_to(self)
            
    @property
    def parent(self):
        return self._parent
            
    def __contains__(self, name_or_node):
        
        if isstr(name_or_node):
            for node in self.nodes:
                if name_or_node == node.name:
                    return True
            else:
                return False
        else:
            return name_or_node in self.nodes
    
    @property
    def qualname(self):
        if not self.name:
            return self._passthrough_qualname()
        
        if self._parent:
            parent_qualname = self._parent.qualname
            if parent_qualname:
                return parent_qualname + '.' + self.name
            else:
                return self.name
        else:
            return self.name
        
    def _passthrough_qualname(self):
        if self._parent:
            return self._parent.qualname
        else:
            return ''
        
    def get_ref_qualname_possibilities(self, ref):
        """ Returns a list of possible qualname references the user might be
        trying to reach. They will be tested in order and the first match 
        will be used. """
        if ref.startswith('.'):
            return [self._parent.qualname + '.' + ref] # sibling ref
        else:
            return [ref] # absolute ref
        
    def attach_to(self, parent):
        self._parent = parent
        parent.nodes.append(self)
        
        if self._creator_interface:
            self._creator_interface.hook_node_attach_to(self, parent) 
            self._creator_interface.hook_child_attach(parent, self) 
        
    def iter_children(self):
        yield self
        for topnode in self.nodes:
            for subnode in topnode.iter_children():
                yield subnode
         



from collections import namedtuple, Sequence
from ..util import yaml_loader

Net = namedtuple('Net', 'refs')

def flatten_string_list(array):
    new_array = []
    for item in array:
        if isinstance(item, Sequence) and not isinstance(item, str):
            new_array += item
        else:
            new_array.append(item)
            
    return new_array

def load_nets(nets_array, nets=None):
    if nets is None:
        nets = {}
        
    
    
    for net_str in flatten_string_list(nets_array):
        refs = _get_net_refs(net_str)

        net = None
        for ref in refs:
            net = nets.get(ref, net)
       
        if net:
            net.refs.update(refs)
        else:
            net = Net(set(refs))
       
        for ref in refs:
            nets[ref] = net

    return nets

ALIAS_SEPARATOR = ','


def _get_net_refs(net):
    if isinstance(net, str):
        refered_nets = net.split(ALIAS_SEPARATOR)
        return [ref.strip() for ref in refered_nets] # Strip removes white space around separators
    else:
        return net.refs
    
def first_alias(net):
    refs = _get_net_refs(net)
    return refs[0]

def load_nets_yml_file(filename):
    return load_nets_dict(yaml_loader.load_yml_file(filename))

def load_nets_yml(yml_string):
    return load_nets_dict(yaml_loader.load_yml(yml_string))

def load_nets_dict(content):
    return load_nets(content['nets'])